@@ -26,24 +26,8 @@
 extern "C" {
 #endif
 
-/**
- * SECTION:timeserver
- * @title: timeserver premitives
- * @short_description: Functions for handling time servers (including NTP)
- */
-
-int connman_timeserver_append(const char *server);
-int connman_timeserver_remove(const char *server);
-
-struct connman_timeserver_driver {
-	const char *name;
-	int priority;
-	int (*append) (const char *server);
-	int (*remove) (const char *server);
-};
-
-int connman_timeserver_driver_register(struct connman_timeserver_driver *driver);
-void connman_timeserver_driver_unregister(struct connman_timeserver_driver *driver);
+int __connman_timeserver_system_append(const char *server);
+int __connman_timeserver_system_remove(const char *server);
 
 #ifdef __cplusplus
 }
